package com.sun.xml.internal.bind.v2.runtime.property;

import com.sun.xml.internal.bind.v2.model.core.PropertyKind;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Loader;
import java.lang.reflect.Type;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.XsiNilLoader;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.DefaultValueLoaderDecorator;
import java.lang.reflect.Modifier;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeTypeInfo;
import com.sun.xml.internal.bind.v2.model.core.TypeRef;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.ChildLoader;
import com.sun.xml.internal.bind.v2.util.QNameMap;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import org.xml.sax.SAXException;
import javax.xml.bind.JAXBElement;
import com.sun.xml.internal.bind.v2.runtime.XMLSerializer;
import com.sun.xml.internal.bind.api.AccessorException;
import com.sun.xml.internal.bind.v2.runtime.JaxBeanInfo;
import java.util.Iterator;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeTypeRef;
import java.util.HashMap;
import com.sun.xml.internal.bind.v2.runtime.JAXBContextImpl;
import com.sun.xml.internal.bind.v2.runtime.Name;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeElementPropertyInfo;
import java.util.Map;
import javax.xml.namespace.QName;
import com.sun.xml.internal.bind.v2.runtime.reflect.Accessor;

final class SingleElementNodeProperty<BeanT, ValueT> extends PropertyImpl<BeanT>
{
    private final Accessor<BeanT, ValueT> acc;
    private final boolean nillable;
    private final QName[] acceptedElements;
    private final Map<Class, TagAndType> typeNames;
    private RuntimeElementPropertyInfo prop;
    private final Name nullTagName;
    
    public SingleElementNodeProperty(JAXBContextImpl context, RuntimeElementPropertyInfo prop) {
        super(context, prop);
        this.typeNames = new HashMap();
        this.acc = prop.getAccessor().optimize(context);
        this.prop = prop;
        QName nt = null;
        boolean nil = false;
        this.acceptedElements = new QName[prop.getTypes().size()];
        for (int i = 0; i < this.acceptedElements.length; ++i) {
            this.acceptedElements[i] = ((RuntimeTypeRef)prop.getTypes().get(i)).getTagName();
        }
        for (RuntimeTypeRef e : prop.getTypes()) {
            JaxBeanInfo beanInfo = context.getOrCreate(e.getTarget());
            if (nt == null) {
                nt = e.getTagName();
            }
            this.typeNames.put(beanInfo.jaxbType, new TagAndType(context.nameBuilder.createElementName(e.getTagName()), beanInfo));
            nil |= e.isNillable();
        }
        this.nullTagName = context.nameBuilder.createElementName(nt);
        this.nillable = nil;
    }
    
    @Override
    public void wrapUp() {
        super.wrapUp();
        this.prop = null;
    }
    
    @Override
    public void reset(BeanT bean) throws AccessorException {
        this.acc.set(bean, null);
    }
    
    @Override
    public String getIdValue(BeanT beanT) {
        return null;
    }
    
    @Override
    public void serializeBody(BeanT o, XMLSerializer w, Object outerPeer) throws SAXException, AccessorException, IOException, XMLStreamException {
        ValueT v = this.acc.get(o);
        if (v != null) {
            Class<Object> vtype = v.getClass();
            TagAndType tt = (TagAndType)this.typeNames.get(vtype);
            if (tt == null) {
                for (Map.Entry<Class, TagAndType> e : this.typeNames.entrySet()) {
                    if (((Class)e.getKey()).isAssignableFrom(vtype)) {
                        tt = (TagAndType)e.getValue();
                        break;
                    }
                    continue;
                }
            }
            boolean addNilDecl = o instanceof JAXBElement && ((JAXBElement)o).isNil();
            if (tt == null) {
                w.startElement(((TagAndType)this.typeNames.values().iterator().next()).tagName, null);
                w.childAsXsiType(v, this.fieldName, w.grammar.getBeanInfo(Object.class), addNilDecl && this.nillable);
            }
            else {
                w.startElement(tt.tagName, null);
                w.childAsXsiType(v, this.fieldName, tt.beanInfo, addNilDecl && this.nillable);
            }
            w.endElement();
        }
        else if (this.nillable) {
            w.startElement(this.nullTagName, null);
            w.writeXsiNilTrue();
            w.endElement();
        }
    }
    
    @Override
    public void buildChildElementUnmarshallers(UnmarshallerChain chain, QNameMap<ChildLoader> handlers) {
        JAXBContextImpl context = chain.context;
        for (TypeRef<Type, Class> e : this.prop.getTypes()) {
            JaxBeanInfo bi = context.getOrCreate((RuntimeTypeInfo)e.getTarget());
            Loader l = bi.getLoader(context, !Modifier.isFinal(bi.jaxbType.getModifiers()));
            if (e.getDefaultValue() != null) {
                l = new DefaultValueLoaderDecorator(l, e.getDefaultValue());
            }
            if (this.nillable || chain.context.allNillable) {
                l = new XsiNilLoader.Single(l, this.acc);
            }
            handlers.put(e.getTagName(), new ChildLoader(l, this.acc));
        }
    }
    
    @Override
    public PropertyKind getKind() {
        return PropertyKind.ELEMENT;
    }
    
    @Override
    public Accessor getElementPropertyAccessor(String nsUri, String localName) {
        QName[] acceptedElements = this.acceptedElements;
        int length = acceptedElements.length;
        int i = 0;
        while (i < length) {
            QName n = acceptedElements[i];
            if (n.getNamespaceURI().equals(nsUri) && n.getLocalPart().equals(localName)) {
                return this.acc;
            }
            ++i;
        }
        return null;
    }
}
